# -*- coding: utf-8 -*-


# PyMeeus: Python module implementing astronomical algorithms.
# Copyright (C) 2018  Dagoberto Salazar
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import sin, cos, sqrt, asin, atan2

from pymeeus.Angle import Angle
from pymeeus.Epoch import Epoch, JDE2000
from pymeeus.Sun import Sun


"""
.. module:: Pluto
   :synopsis: Class to model Pluto minor planet
   :license: GNU Lesser General Public License v3 (LGPLv3)

.. moduleauthor:: Dagoberto Salazar
"""


PLUTO_ARGUMENT = [
    (0.0, 0.0, 1.0),
    (0.0, 0.0, 2.0),
    (0.0, 0.0, 3.0),
    (0.0, 0.0, 4.0),
    (0.0, 0.0, 5.0),
    (0.0, 0.0, 6.0),
    (0.0, 1.0, -1.0),
    (0.0, 1.0, 0.0),
    (0.0, 1.0, 1.0),
    (0.0, 1.0, 2.0),
    (0.0, 1.0, 3.0),
    (0.0, 2.0, -2.0),
    (0.0, 2.0, -1.0),
    (0.0, 2.0, 0.0),
    (1.0, -1.0, 0.0),
    (1.0, -1.0, 1.0),
    (1.0, 0.0, -3.0),
    (1.0, 0.0, -2.0),
    (1.0, 0.0, -1.0),
    (1.0, 0.0, 0.0),
    (1.0, 0.0, 1.0),
    (1.0, 0.0, 2.0),
    (1.0, 0.0, 3.0),
    (1.0, 0.0, 4.0),
    (1.0, 1.0, -3.0),
    (1.0, 1.0, -2.0),
    (1.0, 1.0, -1.0),
    (1.0, 1.0, 0.0),
    (1.0, 1.0, 1.0),
    (1.0, 1.0, 3.0),
    (2.0, 0.0, -6.0),
    (2.0, 0.0, -5.0),
    (2.0, 0.0, -4.0),
    (2.0, 0.0, -3.0),
    (2.0, 0.0, -2.0),
    (2.0, 0.0, -1.0),
    (2.0, 0.0, 0.0),
    (2.0, 0.0, 1.0),
    (2.0, 0.0, 2.0),
    (2.0, 0.0, 3.0),
    (3.0, 0.0, -2.0),
    (3.0, 0.0, -1.0),
    (3.0, 0.0, 0.0)
]
"""This table contains Pluto's argument coefficients according to Table 37.A in
Meeus' book, page 265."""


PLUTO_LONGITUDE = [
    (-19799805.0, 19850055.0),
    (897144.0, -4954829.0),
    (611149.0, 1211027.0),
    (-341243.0, -189585.0),
    (129287.0, -34992.0),
    (-38164.0, 30893.0),
    (20442.0, -9987.0),
    (-4063.0, -5071.0),
    (-6016.0, -3336.0),
    (-3956.0, 3039.0),
    (-667.0, 3572.0),
    (1276.0, 501.0),
    (1152.0, -917.0),
    (630.0, -1277.0),
    (2571.0, -459.0),
    (899.0, -1449.0),
    (-1016.0, 1043.0),
    (-2343.0, -1012.0),
    (7042.0, 788.0),
    (1199.0, -338.0),
    (418.0, -67.0),
    (120.0, -274.0),
    (-60.0, -159.0),
    (-82.0, -29.0),
    (-36.0, -29.0),
    (-40.0, 7.0),
    (-14.0, 22.0),
    (4.0, 13.0),
    (5.0, 2.0),
    (-1.0, 0.0),
    (2.0, 0.0),
    (-4.0, 5.0),
    (4.0, -7.0),
    (14.0, 24.0),
    (-49.0, -34.0),
    (163.0, -48.0),
    (9.0, -24.0),
    (-4.0, 1.0),
    (-3.0, 1.0),
    (1.0, 3.0),
    (-3.0, -1.0),
    (5.0, -3.0),
    (0.0, 0.0)
]
"""This table contains the periodic terms to compute Pluto's heliocentric
longitude according to Table 37.A in Meeus' book, page 265"""


PLUTO_LATITUDE = [
    (-5452852.0, -14974862),
    (3527812.0, 1672790.0),
    (-1050748.0, 327647.0),
    (178690.0, -292153.0),
    (18650.0, 100340.0),
    (-30697.0, -25823.0),
    (4878.0, 11248.0),
    (226.0, -64.0),
    (2030.0, -836.0),
    (69.0, -604.0),
    (-247.0, -567.0),
    (-57.0, 1.0),
    (-122.0, 175.0),
    (-49.0, -164.0),
    (-197.0, 199.0),
    (-25.0, 217.0),
    (589.0, -248.0),
    (-269.0, 711.0),
    (185.0, 193.0),
    (315.0, 807.0),
    (-130.0, -43.0),
    (5.0, 3.0),
    (2.0, 17.0),
    (2.0, 5.0),
    (2.0, 3.0),
    (3.0, 1.0),
    (2.0, -1.0),
    (1.0, -1.0),
    (0.0, -1.0),
    (0.0, 0.0),
    (0.0, -2.0),
    (2.0, 2.0),
    (-7.0, 0.0),
    (10.0, -8.0),
    (-3.0, 20.0),
    (6.0, 5.0),
    (14.0, 17.0),
    (-2.0, 0.0),
    (0.0, 0.0),
    (0.0, 0.0),
    (0.0, 1.0),
    (0.0, 0.0),
    (1.0, 0.0)
]
"""This table contains the periodic terms to compute Pluto's heliocentric
latitude according to Table 37.A in Meeus' book, page 265"""


PLUTO_RADIUS_VECTOR = [
    (66865439.0, 68951812.0),
    (-11827535.0, -332538.0),
    (1593179.0, -1438890.0),
    (-18444.0, 483220.0),
    (-65977.0, -85431.0),
    (31174.0, -6032.0),
    (-5794.0, 22161.0),
    (4601.0, 4032.0),
    (-1729.0, 234.0),
    (-415.0, 702.0),
    (239.0, 723.0),
    (67.0, -67.0),
    (1034.0, -451.0),
    (-129.0, 504.0),
    (480.0, -231.0),
    (2.0, -441.0),
    (-3359.0, 265.0),
    (7856.0, -7832.0),
    (36.0, 45763.0),
    (8663.0, 8547.0),
    (-809.0, -769.0),
    (263.0, -144.0),
    (-126.0, 32.0),
    (-35.0, -16.0),
    (-19.0, -4.0),
    (-15.0, 8.0),
    (-4.0, 12.0),
    (5.0, 6.0),
    (3.0, 1.0),
    (6.0, -2.0),
    (2.0, 2.0),
    (-2.0, -2.0),
    (14.0, 13.0),
    (-63.0, 13.0),
    (136.0, -236.0),
    (273.0, 1065.0),
    (251.0, 149.0),
    (-25.0, -9.0),
    (9.0, -2.0),
    (-8.0, 7.0),
    (2.0, -10.0),
    (19.0, 35.0),
    (10.0, 3.0)
]
"""This table contains the periodic terms to compute Pluto's heliocentric
radius vector according to Table 37.A in Meeus' book, page 265"""


class Pluto(object):
    """
    Class Pluto models that minor planet.
    """

    @staticmethod
    def geometric_heliocentric_position(epoch):
        """This method computes the geometric heliocentric position of planet
        Pluto for a given epoch.

        :param epoch: Epoch to compute Pluto position, as an Epoch object
        :type epoch: :py:class:`Epoch`

        :returns: A tuple with the heliocentric longitude and latitude (as
            :py:class:`Angle` objects), and the radius vector (as a float,
            in astronomical units), in that order
        :rtype: tuple
        :raises: TypeError if input value is of wrong type.
        :raises: ValueError if input epoch outside the 1885-2099 range.

        >>> epoch = Epoch(1992, 10, 13.0)
        >>> l, b, r = Pluto.geometric_heliocentric_position(epoch)
        >>> print(round(l, 5))
        232.74071
        >>> print(round(b, 5))
        14.58782
        >>> print(round(r, 6))
        29.711111
        """

        # First check that input value is of correct types
        if not isinstance(epoch, Epoch):
            raise TypeError("Invalid input type")
        # Check that the input epoch is within valid range
        y = epoch.year()
        if y < 1885.0 or y > 2099.0:
            raise ValueError("Epoch outside the 1885-2099 range")
        t = (epoch - JDE2000) / 36525.0
        jj = 34.35 + 3034.9057 * t
        ss = 50.08 + 1222.1138 * t
        pp = 238.96 + 144.96 * t
        # Compute the arguments
        corr_lon = 0.0
        corr_lat = 0.0
        corr_rad = 0.0
        for n, argument in enumerate(PLUTO_ARGUMENT):
            iii, jjj, kkk = argument
            alpha = Angle(iii * jj + jjj * ss + kkk * pp).to_positive()
            alpha = alpha.rad()
            sin_a = sin(alpha)
            cos_a = cos(alpha)
            a_lon, b_lon = PLUTO_LONGITUDE[n]
            corr_lon += a_lon * sin_a + b_lon * cos_a
            a_lat, b_lat = PLUTO_LATITUDE[n]
            corr_lat += a_lat * sin_a + b_lat * cos_a
            a_rad, b_rad = PLUTO_RADIUS_VECTOR[n]
            corr_rad += a_rad * sin_a + b_rad * cos_a
        # The coefficients in the tables were scaled up. Let's scale them down
        corr_lon /= 1000000.0
        corr_lat /= 1000000.0
        corr_rad /= 10000000.0
        lon = Angle(238.958116 + 144.96 * t + corr_lon)
        lat = Angle(-3.908239 + corr_lat)
        radius = 40.7241346 + corr_rad
        return lon, lat, radius

    @staticmethod
    def geocentric_position(epoch):
        """This method computes the geocentric position of Pluto (right
        ascension and declination) for the given epoch, for the standard
        equinox J2000.0.

        :param epoch: Epoch to compute geocentric position, as an Epoch object
        :type epoch: :py:class:`Epoch`

        :returns: A tuple containing the right ascension and the declination as
            Angle objects
        :rtype: tuple
        :raises: TypeError if input value is of wrong type.
        :raises: ValueError if input epoch outside the 1885-2099 range.

        >>> epoch = Epoch(1992, 10, 13.0)
        >>> ra, dec = Pluto.geocentric_position(epoch)
        >>> print(ra.ra_str(n_dec=1))
        15h 31' 43.7''
        >>> print(dec.dms_str(n_dec=0))
        -4d 27' 29.0''
        """

        # First check that input value is of correct types
        if not isinstance(epoch, Epoch):
            raise TypeError("Invalid input type")
        # Check that the input epoch is within valid range
        y = epoch.year()
        if y < 1885.0 or y > 2099.0:
            raise ValueError("Epoch outside the 1885-2099 range")
        # Compute the heliocentric position of Pluto
        ll, b, r = Pluto.geometric_heliocentric_position(epoch)
        # Change angles to radians
        ll = ll.rad()
        b = b.rad()
        # Values corresponding to obliquity of ecliptic (epsilon) for J2000.0
        sine = 0.397777156
        cose = 0.917482062
        x = r * cos(ll) * cos(b)
        y = r * (sin(ll) * cos(b) * cose - sin(b) * sine)
        z = r * (sin(ll) * cos(b) * sine + sin(b) * cose)
        # Compute Sun's J2000.0 rectacngular coordinates
        xs, ys, zs = Sun.rectangular_coordinates_j2000(epoch)
        # Compute auxiliary quantities
        xi = x + xs
        eta = y + ys
        zeta = z + zs
        # Compute Pluto's distance to Earth
        delta = sqrt(xi * xi + eta * eta + zeta * zeta)
        # Get the light-time difference
        tau = 0.0057755183 * delta
        # Repeat the computations using the light-time correction
        ll, b, r = Pluto.geometric_heliocentric_position(epoch - tau)
        # Change angles to radians
        ll = ll.rad()
        b = b.rad()
        x = r * cos(ll) * cos(b)
        y = r * (sin(ll) * cos(b) * cose - sin(b) * sine)
        z = r * (sin(ll) * cos(b) * sine + sin(b) * cose)
        # Compute auxiliary quantities
        xi = x + xs
        eta = y + ys
        zeta = z + zs
        # Compute Pluto's distance to Earth
        delta = sqrt(xi * xi + eta * eta + zeta * zeta)
        # Compute right ascension and declination
        alpha = Angle(atan2(eta, xi), radians=True)
        dec = Angle(asin(zeta / delta), radians=True)
        return alpha.to_positive(), dec


def main():

    # Let's define a small helper function
    def print_me(msg, val):
        print("{}: {}".format(msg, val))

    # Let's show some uses of Pluto class
    print("\n" + 35 * "*")
    print("*** Use of Pluto class")
    print(35 * "*" + "\n")

    # Let's now compute the heliocentric position for a given epoch
    epoch = Epoch(1992, 10, 13.0)
    lon, lat, r = Pluto.geometric_heliocentric_position(epoch)
    print_me("Geometric Heliocentric Longitude", lon.to_positive())
    print_me("Geometric Heliocentric Latitude", lat)
    print_me("Radius vector", r)

    print("")

    # Compute the geocentric position for 1992/10/13:
    epoch = Epoch(1992, 10, 13.0)
    ra, dec = Pluto.geocentric_position(epoch)
    print_me("Right ascension", ra.ra_str(n_dec=1))
    print_me("Declination", dec.dms_str(n_dec=1))


if __name__ == "__main__":

    main()



# -*- coding: utf-8 -*-


# PyMeeus: Python module implementing astronomical algorithms.
# Copyright (C) 2018  Dagoberto Salazar
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import sqrt, degrees

from pymeeus.base import TOL
from pymeeus.Angle import Angle


"""
.. module:: Interpolation
   :synopsis: Basic interpolation routines
   :license: GNU Lesser General Public License v3 (LGPLv3)

.. moduleauthor:: Dagoberto Salazar
"""


class Interpolation(object):
    """
    Class Interpolation deals with finding intermediate values to those given
    in a table.

    Besides the basic interpolation, this class also provides methods to find
    the extrema (maximum or minimum value) in a given interval (if present),
    and it also has methods to find roots (the values of the argument 'x' for
    which the function 'y' becomes zero).

    Please note that it seems that Meeus uses the Bessel interpolation method
    (Chapter 3). However, this class uses the Newton interpolation method
    because it is (arguably) more flexible regarding the number of entries
    provided in the interpolation table.

    The constructor takes pairs of (x, y) values from the table of interest.
    These pairs of values can be given as a sequence of int/floats, tuples,
    lists or Angles. It is also possible to provide an Interpolation object to
    the constructor in order to get a copy.

    .. note:: When using Angles, be careful with the 360-to-0 discontinuity.

    If a sequence of int, floats or Angles is given, the values in the odd
    positions are considered to belong to the 'x' set, while the values in the
    even positions belong to the 'y' set. If only one tuple or list is
    provided, it is assumed that it is the 'y' set, and the 'x' set is build
    from 0 onwards with steps of length 1.

    Please keep in mind that a minimum of two data pairs are needed in order to
    carry out any interpolation. If only one value pair is provided, a
    ValueError exception will be raised.
    """

    def __init__(self, *args):
        """Interpolation constructor.

        This takes pairs of (x, y) values from the table of interest. These
        pairs of values can be given as a sequence of int/floats, tuples, lists
        or Angles. It is also possible to provide an Interpolation object to
        the constructor in order to get a copy.

        .. note:: When using Angles, be careful with the 360-to-0 discontinuity

        If a sequence of int, floats or Angles is given, the values in the odd
        positions are considered to belong to the 'x' set, while the values in
        the even positions belong to the 'y' set. If only one tuple or list is
        provided, it is assumed that it is the 'y' set, and the 'x' set is
        build from 0 onwards with steps of length 1.

        Please keep in mind that a minimum of two data pairs are needed in
        order to carry out any interpolation. If only one value pair is
        provided, a ValueError exception will be raised.

        :param args: Input tabular values, or another Interpolation object.
        :type args: int, float, list, tuple, :py:class:`Angle`,
           :py:class:`Interpolation`

        :returns: Interpolation object.
        :rtype: :py:class:`Interpolation`
        :raises: ValueError if not enough input data pairs are provided.
        :raises: TypeError if input values are of wrong type.

        >>> i = Interpolation([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
        >>> print(i._x)
        [1, 2, 3, 4, 5, 6]
        >>> print(i._y)
        [2, 4, 6, 8, 10, 12]
        >>> j = Interpolation([3, -8, 1, 12, 2, 5, 8])
        >>> print(j._x)
        [0, 1, 2, 3, 4, 5, 6]
        >>> print(j._y)
        [3, -8, 1, 12, 2, 5, 8]
        >>> k = Interpolation(3, -8, 1, 12, 2, 5, 8)
        >>> print(k._x)
        [1, 2, 3]
        >>> print(k._y)
        [12, 5, -8]
        """

        self._x = []
        self._y = []
        self._table = []
        self._tol = TOL
        self.set(*args)  # Let's use 'set()' method to handle the setup

    def _order_points(self):
        """Method to put the data points in ascending order w.r.t. 'x'."""

        # Let's work with copies of the original lists
        x = list(self._x)
        y = list(self._y)

        xnew = []
        ynew = []
        xmax = max(x) + 1.0
        for _ in range(len(x)):
            # Get the index of the minimum value
            imin = x.index(min(x))
            # Append the current minimum value to the new 'x' list
            xnew.append(x[imin])
            # Store the *position* of the current minimum value to new 'y' list
            ynew.append(imin)
            # The current minimum value will no longer be the minimum
            x[imin] = xmax

        # In the new 'y' list, substitute the positions by the real values
        for i in range(len(x)):
            ynew[i] = y[ynew[i]]

        # Store the results in the corresponding fields
        self._x = xnew
        self._y = ynew

    def set(self, *args):
        """Method used to define the value pairs of Interpolation object.

        This takes pairs of (x, y) values from the table of interest. These
        pairs of values can be given as a sequence of int/floats, tuples,
        lists, or Angles. It is also possible to provide an Interpolation
        object to this method in order to get a copy.

        .. note:: When using Angles, be careful with the 360-to-0 discontinuity

        If a sequence of int, floats or Angles is given, the values in the odd
        positions are considered to belong to the 'x' set, while the values in
        the even positions belong to the 'y' set. If only one tuple or list is
        provided, it is assumed that it is the 'y' set, and the 'x' set is
        build from 0 onwards with steps of length 1.

        Please keep in mind that a minimum of two data pairs are needed in
        order to carry out any interpolation. If only one value pair is
        provided, a ValueError exception will be raised.

        :param args: Input tabular values, or another Interpolation object.
        :type args: int, float, list, tuple, :py:class:`Angle`

        :returns: None.
        :rtype: None
        :raises: ValueError if not enough input data pairs are provided.
        :raises: TypeError if input values are of wrong type.

        >>> i = Interpolation()
        >>> i.set([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
        >>> print(i._x)
        [1, 2, 3, 4, 5, 6]
        >>> print(i._y)
        [2, 4, 6, 8, 10, 12]
        >>> j = Interpolation()
        >>> j.set([3, -8, 1, 12, 2, 5, 8])
        >>> print(j._x)
        [0, 1, 2, 3, 4, 5, 6]
        >>> print(j._y)
        [3, -8, 1, 12, 2, 5, 8]
        >>> k = Interpolation(3, -8, 1, 12, 2, 5, 8)
        >>> print(k._x)
        [1, 2, 3]
        >>> print(k._y)
        [12, 5, -8]
        >>> m = Interpolation(k)
        >>> print(m._x)
        [1, 2, 3]
        >>> print(m._y)
        [12, 5, -8]
        """

        # Clean up the internal data tables
        self._x = []
        self._y = []
        self._table = []
        # If no arguments are given, return. Internal data tables are empty
        if len(args) == 0:
            return
        # If we have only one argument, it can be a single value or tuple/list
        elif len(args) == 1:
            if isinstance(args[0], Interpolation):  # Copy constructor
                self._x = args[0]._x
                self._y = args[0]._y
                self._table = args[0]._table
                self._tol = args[0]._tol
                return
            elif isinstance(args[0], (int, float, Angle)):
                # Insuficient data to interpolate. Raise ValueError exception
                raise ValueError("Invalid number of input values")
            elif isinstance(args[0], (list, tuple)):
                seq = args[0]
                if len(seq) < 2:
                    raise ValueError("Invalid number of input values")
                else:
                    # Read input values into 'y', and create 'x'
                    i = 0
                    for value in seq:
                        self._x.append(i)
                        self._y.append(value)
                        i += 1
            else:
                raise TypeError("Invalid input value")
        elif len(args) == 2:
            if isinstance(args[0], (int, float, Angle)) or isinstance(
                args[1], (int, float, Angle)
            ):
                # Insuficient data to interpolate. Raise ValueError exception
                raise ValueError("Invalid number of input values")
            elif isinstance(args[0], (list, tuple)) and isinstance(
                args[1], (list, tuple)
            ):
                x = args[0]
                y = args[1]
                # Check if they have the same length. If not, make them equal
                length_min = min(len(x), len(y))
                x = x[:length_min]
                y = y[:length_min]
                if len(x) < 2 or len(y) < 2:
                    raise ValueError("Invalid number of input values")
                else:
                    # Read input values into 'x' and 'y'
                    for xval, yval in zip(x, y):
                        self._x.append(xval)
                        self._y.append(yval)
            else:
                raise TypeError("Invalid input value")
        elif len(args) == 3:
            # In this case, no combination of input values is valid
            raise ValueError("Invalid number of input values")
        else:
            # If there is an odd number of arguments, drop the last one
            if len(args) % 2 != 0:
                args = args[:-1]
            # Check that all the arguments are ints, floats or Angles
            all_numbers = True
            for arg in args:
                all_numbers = (all_numbers
                               and isinstance(arg, (int, float, Angle)))
            # If any of the values failed the test, raise an exception
            if not all_numbers:
                raise TypeError("Invalid input value")
            # Now, extract the data: Odds are x's, evens are y's
            for i in range(int(len(args) / 2.0)):
                self._x.append(args[2 * i])
                self._y.append(args[2 * i + 1])
        # After self._x is found, confirm that x's are different to each other
        for i in range(len(self._x) - 1):
            for k in range(i + 1, len(self._x)):
                if abs(self._x[i] - self._x[k]) < self._tol:
                    raise ValueError("Invalid input: Values in 'x' are equal")
        # Order the data points if needed
        self._order_points()
        # Create table containing Newton coefficientes, only if values given
        if len(self._x) > 0:
            self._compute_table()

    def __str__(self):
        """Method used when trying to print the object.

        :returns: Internal tabular values as strings.
        :rtype: string

        >>> i = Interpolation([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
        >>> print(i)
        X: [1, 2, 3, 4, 5, 6]
        Y: [2, 4, 6, 8, 10, 12]
        """

        xstr = "X: " + str(self._x) + "\n"
        ystr = "Y: " + str(self._y)
        return xstr + ystr

    def __repr__(self):
        """Method providing the 'official' string representation of the object.
        It provides a valid expression that could be used to recreate the
        object.

        :returns: As string with a valid expression to recreate the object
        :rtype: string

        >>> i = Interpolation([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
        >>> repr(i)
        'Interpolation([1, 2, 3, 4, 5, 6], [2, 4, 6, 8, 10, 12])'
        """

        return "{}({}, {})".format(self.__class__.__name__, self._x, self._y)

    def __len__(self):
        """This method returns the number of interpolation points (x, y, pairs)
        stored in this :class:`Interpolation` object.

        :returns: Number of interpolation points internally stored
        :rtype: int

        >>> i = Interpolation([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
        >>> len(i)
        6
        """

        return len(self._x)

    def get_tolerance(self):
        """Gets the internal tolerance value used for comparisons.

        .. note:: The default tolerance value is TOL.

        :returns: Internal tolerance.
        :rtype: float
        """

        return self._tol

    def set_tolerance(self, tol):
        """Changes the internal tolerance value used for comparisons.

        :param tol: New tolerance value.
        :type tol: int, float

        :returns: None
        :rtype: None
        """

        self._tol = tol
        return

    def _compute_table(self):
        """Method to compute coefficients of Newton interpolation method."""

        for i in range(len(self._x)):
            self._table.append(self._newton_diff(0, i))

    def _newton_diff(self, start, end):
        """Auxiliary method to compute the elements of the Newton table.

        :param start: Starting index
        :type start: int
        :param end: Ending index
        :type end: int

        :returns: Resulting value of the element of the Newton table.
        :rtype: float
        """

        if abs(end - start) < self._tol:
            val = self._y[start]
        else:
            x = list(self._x)  # Let's make a copy, just in case
            val = (self._newton_diff(start, end - 1)
                   - self._newton_diff(start + 1, end)) / (x[start] - x[end])

        return val

    def __call__(self, x):
        """Method to interpolate the function at a given 'x'.

        :param x: Point where the interpolation will be carried out.
        :type x: int, float, :py:class:`Angle`

        :returns: Resulting value of the interpolation.
        :rtype: float
        :raises: ValueError if input value is outside of interpolation range.
        :raises: TypeError if input value is of wrong type.

        >>> i = Interpolation([7, 8, 9], [0.884226, 0.877366, 0.870531])
        >>> y = round(i(8.18125), 6)
        >>> print(y)
        0.876125
        """

        # Check if input value is of correct type
        if isinstance(x, (int, float, Angle)):
            # Check if 'x' already belongs to the data table
            for i in range(len(self._x)):
                if abs(x - self._x[i]) < self._tol:
                    return self._y[i]  # We don't need to look further
            # Check if Newton coefficients table is not empty
            if len(self._table) == 0:
                raise RuntimeError("Internal table is empty. Use set().")
            # Check that x is within interpolation table values
            if x < self._x[0] or x > self._x[-1]:
                raise ValueError("Input value outside of interpolation range.")
            # Horner's method is used to efficiently compute the result
            val = self._table[-1]
            for i in range(len(self._table) - 1, 0, -1):
                val = self._table[i - 1] + (x - self._x[i - 1]) * val

            return val
        else:
            raise TypeError("Invalid input value")

    def derivative(self, x):
        """Method to compute the derivative from interpolation polynomial.

        :param x: Point where the interpolation derivative will be carried out.
        :type x: int, float, :py:class:`Angle`

        :returns: Resulting value of the interpolation derivative.
        :rtype: float
        :raises: ValueError if input value is outside of interpolation range.
        :raises: TypeError if input value is of wrong type.

        >>> m = Interpolation([-1.0, 0.0, 1.0], [-2.0, 3.0, 2.0])
        >>> m.derivative(-1.0)
        8.0
        >>> m.derivative(0.5)
        -1.0
        """

        # Check if input value is of correct type
        if isinstance(x, (int, float, Angle)):
            # Check that x is within interpolation table values
            if x < self._x[0] or x > self._x[-1]:
                raise ValueError("Input value outside of interpolation range.")
            # If we only have two interpolation points, derivative is simple
            if len(self._x) == 2:
                return (self._y[1] - self._y[0]) / (self._x[1] - self._x[0])
            else:
                res = self._table[1]
                for k in range(len(self._table) - 1, 1, -1):
                    val = 0.0
                    for j in range(k):
                        s = 1.0
                        for i in range(k):
                            if i != j:
                                s *= x - self._x[i]
                        val += s
                    res += val * self._table[k]
                return res
        else:
            raise TypeError("Invalid input value")

    def root(self, xl=0, xh=0, max_iter=1000):
        """Method to find the root inside the [xl, xh] range.

        This method applies, in principle, the Newton method to find the root;
        however, if conditions are such that Newton method may not bei properly
        behaving or converging, then it switches to the linear Interpolation
        method.

        If values xl, xh are not given, the limits of the interpolation table
        values will be used.

        .. note:: This method returns a ValueError exception if the
           corresponding yl = f(xl) and yh = f(xh) values have the same sign.
           In that case, the method assumes there is no root in the [xl, xh]
           interval.

        .. note:: If any of the xl, xh values is beyond the limits given by the
           interpolation values, its value will be set to the corresponding
           limit.

        .. note:: If xl == xh (and not zero), a ValueError exception is raised.

        .. note:: If the method doesn't converge within max_iter ierations,
           then a ValueError exception is raised.

        :param xl: Lower limit of interval where the root will be looked for.
        :type xl: int, float, :py:class:`Angle`
        :param xh: Higher limit of interval where the root will be looked for.
        :type xh: int, float, :py:class:`Angle`
        :param max_iter: Maximum number of iterations allowed.
        :type max_iter: int

        :returns: Root of the interpolated function within [xl, xh] interval.
        :rtype: int, float, :py:class:`Angle`
        :raises: ValueError if yl = f(xl), yh = f(xh) have same sign.
        :raises: ValueError if xl == xh.
        :raises: ValueError if maximum number of iterations is exceeded.
        :raises: TypeError if input value is of wrong type.

        >>> m = Interpolation([-1.0, 0.0, 1.0], [-2.0, 3.0, 2.0])
        >>> round(m.root(), 8)
        -0.72075922
        """

        # Get the limits of the interpolation table
        xmin = self._x[0]
        xmax = self._x[-1]
        # Check if input value is of correct type
        if (isinstance(xl, (int, float, Angle))
                and isinstance(xh, (int, float, Angle))):
            # Check if BOTH values are zero
            if xl == 0 and xh == 0:
                xl = xmin
                xh = xmax
            # Check if limits are equal
            if abs(xl - xh) < self._tol:
                raise ValueError("Invalid limits: xl and xh values are equal")
            # Put limits in order. Swap them if necessary
            if xl > xh:
                xl, xh = xh, xl
            # Check limits against interpolation table. Reset if necessary
            if xl < self._x[0]:
                xl = xmin
            if xh < self._x[-1]:
                xh = xmax
            yl = self.__call__(xl)
            yh = self.__call__(xh)
            # Check for a couple special cases
            if abs(yl) < self._tol:
                return xl  # xl is a root
            if abs(yh) < self._tol:
                return xh  # xh is a root
            # Check if signs of ordinates are the same
            if (yl * yh) > 0.0:
                raise ValueError("Invalid interval: Probably no root exists")
            # We are good to go. First option: Newton's root-finding method
            x = (xl + xh) / 2.0  # Start in the middle of interval
            y = self.__call__(x)
            num_iter = 0  # Count the number of iterations
            while abs(y) > self._tol:
                if num_iter >= max_iter:
                    raise ValueError(
                        "Too many iterations: Probably no root\
                                     exists"
                    )
                num_iter += 1
                yp = self.derivative(x)
                # If derivative is too small, switch to linear interpolation
                if abs(yp) < 1e-3:
                    x = (xl * yh - xh * yl) / (yh - yl)
                    y = self.__call__(x)
                else:
                    x = x - y / yp
                    # Check if x is within limits
                    if x < xmin or x > xmax:
                        # Switch to linear interpolation
                        x = (xl * yh - xh * yl) / (yh - yl)
                        y = self.__call__(x)
                    else:
                        y = self.__call__(x)
                if (y * yl) >= 0.0:
                    xl = x
                    yl = y
                else:
                    xh = x
                    yh = y
            return x
        else:
            raise TypeError("Invalid input value")

    def minmax(self, xl=0, xh=0, max_iter=1000):
        """Method to find the minimum or maximum inside the [xl, xh] range.

        Finding the minimum or maximum (extremum) of a function within a given
        interval is akin to find the root of its derivative. Therefore, this
        method creates an interpolation object for the derivative function, and
        calls the root method of that object. See :meth:`root` method for more
        details.

        If values xl, xh are not given, the limits of the interpolation table
        values will be used.

        .. note::

           This method returns a ValueError exception if the corresponding
           derivatives yl' = f'(xl) and yh' = f'(xh) values have the same sign.
           In that case, the method assumes there is no extremum in the
           [xl, xh] interval.

        .. note::

           If any of the xl, xh values is beyond the limits given by the
           interpolation values, its value will be set to the corresponding
           limit.

        .. note:: If xl == xh (and not zero), a ValueError exception is raised.

        .. note:: If the method doesn't converge within max_iter ierations,
           then a ValueError exception is raised.

        :param xl: Lower limit of interval where a extremum will be looked for.
        :type xl: int, float, :py:class:`Angle`
        :param xh: Higher limit of interval where extremum will be looked for.
        :type xh: int, float, :py:class:`Angle`
        :param max_iter: Maximum number of iterations allowed.
        :type max_iter: int

        :returns: Extremum of interpolated function within [xl, xh] interval.
        :rtype: int, float, :py:class:`Angle`
        :raises: ValueError if yl = f(xl), yh = f(xh) have same sign.
        :raises: ValueError if xl == xh.
        :raises: ValueError if maximum number of iterations is exceeded.
        :raises: TypeError if input value is of wrong type.

        >>> m = Interpolation([-1.0, 0.0, 1.0], [-2.0, 3.0, 2.0])
        >>> round(m.minmax(), 8)
        0.33333333
        """

        # Compute the derivatives for the current data
        x = list(self._x)
        y = []
        for xi in x:
            y.append(self.derivative(xi))
        # Create a new Interpolation object
        prime = Interpolation(x, y)
        # Find the root within that object, and return it
        return prime.root(xl, xh, max_iter)


def main():

    # Let's define a small helper function
    def print_me(msg, val):
        print("{}: {}".format(msg, val))

    # Let's now work with the Interpolation class
    print("\n" + 35 * "*")
    print("*** Use of Interpolation class")
    print(35 * "*" + "\n")

    i = Interpolation([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])
    print("i = Interpolation([5, 3, 6, 1, 2, 4, 9], [10, 6, 12, 2, 4, 8])")
    print(i)
    print("NOTE:")
    print("   a. They are ordered in 'x'")
    print("   b. The extra value in 'x' was dropped")

    print("")

    # Use the copy constructor
    print("We can easily make a copy of an Interpolation object")
    j = Interpolation(i)
    print("j = Interpolation(i)")
    print(j)

    print("")

    # Get the number of interpolation points stored
    print_me("Number or interpolation points in 'j'", len(j))  # 6

    print("")

    j = Interpolation([0.0, 1.0, 3.0], [-1.0, -2.0, 2.0])
    print("j = Interpolation([0.0, 1.0, 3.0], [-1.0, -2.0, 2.0])")
    print(j)
    print_me("j(2)", j(2))
    print_me("j(0.5)", j(0.5))
    # Test with a value already in the data table
    print_me("j(1)", j(1))

    print("")

    # We can interpolate Angles too
    k = Interpolation(
        [27.0, 27.5, 28.0, 28.5, 29.0],
        [
            Angle(0, 54, 36.125),
            Angle(0, 54, 24.606),
            Angle(0, 54, 15.486),
            Angle(0, 54, 8.694),
            Angle(0, 54, 4.133),
        ],
    )
    print(
        "k = Interpolation([27.0, 27.5, 28.0, 28.5, 29.0],\n\
                      [Angle(0, 54, 36.125), Angle(0, 54, 24.606),\n\
                       Angle(0, 54, 15.486), Angle(0, 54, 8.694),\n\
                       Angle(0, 54, 4.133)])"
    )

    print_me("k(28.278)", Angle(k(28.278)).dms_str())

    print("")

    m = Interpolation([-1.0, 0.0, 1.0], [-2.0, 3.0, 2.0])
    print("m = Interpolation([-1.0, 0.0, 1.0], [-2.0, 3.0, 2.0])")
    print(m)
    # Get interpolated values
    print_me("m(-0.5)", m(-0.5))
    print_me("m(0.5)", m(0.5))
    # Get derivatives
    print_me("m'(-1.0)", m.derivative(-1.0))
    print_me("m'(-0.5)", m.derivative(-0.5))
    print_me("m'(0.0)", m.derivative(0.0))
    print_me("m'(0.5)", m.derivative(0.5))
    print_me("m'(1.0)", m.derivative(1.0))
    # Get the root within the interval
    print_me("m.root()", m.root())
    # Get the extremum within the interval
    print_me("m.minmax()", m.minmax())

    m = Interpolation(
        [29.43, 30.97, 27.69, 28.11, 31.58, 33.05],
        [
            0.4913598528,
            0.5145891926,
            0.4646875083,
            0.4711658342,
            0.5236885653,
            0.5453707057,
        ],
    )

    print_me("sin(29.5)\t", m(29.5))
    print_me("sin(30.0)\t", m(30.0))
    print_me("sin(30.5)\t", m(30.5))
    # Derivative must be adjusted because degrees were used instead of radians
    print_me("sin'(29.5)\t", degrees(m.derivative(29.5)))
    print_me("sin'(30.0)\t", degrees(m.derivative(30.0)))
    print_me("sqrt(3.0)/2.0\t", sqrt(3.0) / 2.0)
    print_me("sin'(30.5)\t", degrees(m.derivative(30.5)))


if __name__ == "__main__":

    main()



# -*- coding: utf-8 -*-


# PyMeeus: Python module implementing astronomical algorithms.
# Copyright (C) 2018  Dagoberto Salazar
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from math import floor


"""
.. module:: base
   :synopsis: Basic routines and constants used by the pymeeus module
   :license: GNU Lesser General Public License v3 (LGPLv3)

.. moduleauthor:: Dagoberto Salazar
"""


TOL = 1e-10
"""Internal tolerance being used by default"""


def machine_accuracy():
    """This function computes the accuracy of the computer being used.

    This function returns a tuple containing the number of significant bits in
    the mantissa of a floating number, and the number of significant digits in
    a decimal number.

    :returns: Number of significant bits, and of significant digits
    :rtype: tuple
    """

    j = 0.0
    x = 2.0
    while x + 1.0 != x:
        j += 1.0
        x *= 2.0
    return (j, int(j * 0.30103))


def get_ordinal_suffix(ordinal):
    """Method to get the suffix of a given ordinal number, like 1'st',
    2'nd', 15'th', etc.

    :param ordinal: Ordinal number
    :type ordinal: int

    :returns: Suffix corresponding to input ordinal number
    :rtype: str
    :raises: TypeError if input type is invalid.

    >>> get_ordinal_suffix(40)
    'th'
    >>> get_ordinal_suffix(101)
    'st'
    >>> get_ordinal_suffix(2)
    'nd'
    >>> get_ordinal_suffix(19)
    'th'
    >>> get_ordinal_suffix(23)
    'rd'
    """

    if not isinstance(ordinal, (int, float)):
        raise TypeError("Invalid input type")
    else:
        ordinal = int(floor(ordinal))
        unit = ordinal % 10
        if unit == 1 and ordinal != 11:
            return "st"
        elif unit == 2 and ordinal != 12:
            return "nd"
        elif unit == 3 and ordinal != 13:
            return "rd"
        else:
            return "th"


def iint(number):
    """This method behaves in the same way as the **INT()** function described
    by Meeus in his book: Greatest integer which is not greater than number.

    :param number: Number or expresion
    :type number: int, float

    :returns: Greatest integer which is not greater than number
    :rtype: int
    :raises: TypeError if input type is invalid.

    >>> iint(19)
    19
    >>> iint(19.95)
    19
    >>> iint(-2.4)
    -3
    """

    if not isinstance(number, (int, float)):
        raise TypeError("Invalid input type")
    else:
        return int(floor(number))


def main():

    # Let's define a small helper function
    def print_me(msg, val):
        print("{}: {}".format(msg, val))

    # Let's print the tolerance
    print_me("The default value for the tolerance is", TOL)

    # Find the accuracy of this computer
    j, d = machine_accuracy()
    print_me("Number of significant BITS in the mantissa\t", j)
    print_me("Number of significant DIGITS in a decimal number", d)

    print("")

    print_me("The suffix for ordinal 2 is", get_ordinal_suffix(2))
    print_me("The suffix for ordinal 11 is", get_ordinal_suffix(11))
    print_me("The suffix for ordinal 12 is", get_ordinal_suffix(12))
    print_me("The suffix for ordinal 13 is", get_ordinal_suffix(13))
    print_me("The suffix for ordinal 14 is", get_ordinal_suffix(14))
    print_me("The suffix for ordinal 16 is", get_ordinal_suffix(16))
    print_me("The suffix for ordinal 23 is", get_ordinal_suffix(23))


if __name__ == "__main__":

    main()
